/*
**  Copyright 2000-2004 University of Illinois Board of Trustees
**  Copyright 2000-2004 Mark D. Roth
**  All rights reserved.
**
**  download.c - FTP downloading code for fget
**
**  Mark D. Roth <roth@feep.net>
*/

#include <fget.h>

#include <stdio.h>
#include <errno.h>
#include <utime.h>
#include <sys/stat.h>
#include <time.h>

#ifdef STDC_HEADERS
# include <string.h>
# include <stdlib.h>
#endif


static int
fget_download_file(FTP *ftp, char *remotepath, struct ftpstat *fsp,
		   void *localtree)
{
	int i;
	char localpath[MAXPATHLEN], linkbuf[MAXPATHLEN];
	struct stat s;
	struct utimbuf utb;

#ifdef DEBUG
	printf("==> fget_download_file(remotepath=\"%s\", localtree=\"%s\")\n",
	       remotepath, (char *)localtree);
#endif

	strlcpy(localpath, (char *)localtree, sizeof(localpath));
	if (strcmp(remotepath, "/") != 0)
	{
		if (localpath[0] != '\0')
			strlcat(localpath, "/", sizeof(localpath));
		strlcat(localpath, basename(remotepath), sizeof(localpath));
	}

	memset(&s, 0, sizeof(s));
	if (lstat(localpath, &s) == -1 && errno != ENOENT)
	{
		fprintf(stderr, "fget: lstat(\"%s\"): %s\n", localpath,
			strerror(errno));
		return R_ERR;
	}

	if (S_ISDIR(fsp->fs_mode))
	{
		i = fget_download_dir(ftp, remotepath, fsp, localpath, &s);
		if (i != R_FILEOK && i != R_SKIP)
			return i;

		if (BIT_ISSET(options, OPT_RECURSIVE))
		{
			if (verbose)
				printf("fget: checking %s...\n", remotepath);
			if (fget_recursion(ftp, remotepath,
					   fget_download_file,
					   localpath) == R_RETURN)
				return R_RETURN;
			if (BIT_ISSET(options, OPT_DELETELOCAL))
				fget_delete_local(ftp, remotepath, localpath);
		}
	}
	else if (S_ISLNK(fsp->fs_mode))
	{
		i = fget_download_symlink(ftp, remotepath, fsp, localpath, &s,
					  linkbuf, sizeof(linkbuf));
		if (i != R_FILEOK && i != R_SKIP)
			return i;
	}
	else if (S_ISREG(fsp->fs_mode))
	{
		i = fget_download_regfile(ftp, remotepath, fsp, localpath, &s);
		if (i != R_FILEOK && i != R_SKIP)
			return i;
	}
	else
	{
		fprintf(stderr, "fget: unsupported file type (mode == 0%6o)\n",
			(unsigned int)fsp->fs_mode);
		return R_ERR;
	}

	/* update timestamp */
	if (!S_ISLNK(fsp->fs_mode))
	{
		utb.actime = time(NULL);
		utb.modtime = fsp->fs_mtime;
		if (utime(localpath, &utb) == -1)
		{
			fprintf(stderr, "fget: utime(\"%s\"): %s\n",
				localpath, strerror(errno));
			return R_ERR;
		}
	}

	if (BIT_ISSET(options, OPT_SAVEPERMS))
		fget_set_perms(localpath, fsp);

	return R_FILEOK;
}


static int
fget_download_root(FTP *ftp, char *remotepath, char *root,
		   char *cwd, size_t cwdsize)
{
	char rpbuf[MAXPATHLEN], localpath[MAXPATHLEN];
	char tmp[MAXPATHLEN], linkbuf[MAXPATHLEN];
	int i;
	char *thisp, *nextp = rpbuf;
	struct ftpstat fs;
	struct stat s;

#ifdef DEBUG
	printf("==> fget_download_root(ftp=(%s), remotepath=\"%s\", "
	       "root=\"%s\")\n", ftp_get_host(ftp), remotepath, root);
#endif

	if (strcmp(ftp_getcwd(ftp), "/") != 0 || strcmp(remotepath, "/") == 0)
		strlcpy(cwd, ftp_getcwd(ftp), cwdsize);
	else
		cwd[0] = '\0';

	strlcpy(rpbuf, remotepath, sizeof(rpbuf));
	while ((thisp = strsep(&nextp, "/")) != NULL)
	{
		if (*thisp == '\0')
			continue;

		strlcat(cwd, "/", cwdsize);
		strlcat(cwd, thisp, cwdsize);
#ifdef DEBUG
		printf("    fget_download_root(): cwd = \"%s\"\n", cwd);
		printf("    fget_download_root(): nextp = \"%s\"\n", nextp);
#endif
		if (fget_stat(ftp, cwd, &fs) != R_FILEOK)
			return R_RETURN;

		strlcpy(localpath, root, sizeof(localpath));
		strlcat(localpath, cwd, sizeof(localpath));
#ifdef DEBUG
		printf("    fget_download_root(): localpath = \"%s\"\n",
		       localpath);
#endif
		memset(&s, 0, sizeof(s));
		if (lstat(localpath, &s) == -1 && errno != ENOENT)
		{
			fprintf(stderr, "fget: lstat(\"%s\"): %s\n",
				localpath, strerror(errno));
			return R_RETURN;
		}

		if (S_ISDIR(fs.fs_mode))
		{
			i = fget_download_dir(ftp, cwd, &fs, localpath, &s);
			if (i != R_FILEOK && i != R_SKIP)
				return i;
		}
		else if (S_ISLNK(fs.fs_mode))
		{
			i = fget_download_symlink(ftp, cwd, &fs, localpath,
						  &s, linkbuf, sizeof(linkbuf));
			if (i != R_FILEOK && i != R_SKIP)
				return i;

			/* construct new path */
			snprintf(tmp, sizeof(tmp), "%s/%s",
				 dirname(cwd), linkbuf);
			if (nextp != NULL && *nextp != '\0')
			{
				strlcat(tmp, "/", sizeof(tmp));
				strlcat(tmp, nextp, sizeof(tmp));
			}
			if (fget_cleanpath(tmp, linkbuf, sizeof(linkbuf)) == -1)
			{
				fprintf(stderr,
					"fget: fget_cleanpath(\"%s\"): %s\n",
					tmp, strerror(errno));
				return R_ERR;
			}
			return fget_download_root(ftp, linkbuf, root, cwd,
						  cwdsize);
		}
		else if (nextp != NULL && *nextp != '\0')
		{
			fprintf(stderr, "fget: %s: %s\n", cwd,
				strerror(ENOTDIR));
			return R_RETURN;
		}
	}

	return R_FILEOK;
}


int
fget_download(FTP *ftp, char *remotepath, char *changeroot)
{
	struct ftpstat fs;
	char localroot[MAXPATHLEN] = ".", newpath[MAXPATHLEN];
	int retval = 0, freeflag = 0;
	fget_listptr_t lp1, lp2;
	char *cp;

#ifdef DEBUG
	printf("==> fget_download(ftp=(%s), remotepath=\"%s\", "
	       "changeroot=\"%s\")\n",
	       ftp_get_host(ftp), remotepath,
	       (changeroot ? changeroot : "NULL"));
#endif

	if (changeroot)
		strlcpy(localroot, changeroot, sizeof(localroot));
	if (BIT_ISSET(options, OPT_HOSTDIRS))
	{
		strlcat(localroot, "/", sizeof(localroot));
		strlcat(localroot, ftp_get_host(ftp), sizeof(localroot));
	}
	if (fget_mkdirhier(localroot, NULL, NULL) == -1)
	{
		fprintf(stderr, "fget: fget_mkdirhier(\"%s\"): %s\n",
			localroot, strerror(errno));
		return R_RETURN;
	}

	if (BIT_ISSET(options, OPT_HOSTDIRS))
	{
		if (fget_download_root(ftp, remotepath,
				       localroot, newpath,
				       sizeof(newpath)) != R_FILEOK)
			return R_RETURN;
		if (strcmp(newpath, "/") != 0)
		{
			if (newpath[0] != '/')
				strlcat(localroot, "/", sizeof(localroot));
			strlcat(localroot, newpath, sizeof(localroot));
		}
	}
	else
	{
		if (ftp_realpath(ftp, remotepath, newpath,
				 sizeof(newpath)) == -1)
		{
			fprintf(stderr, "fget: ftp_realpath(\"%s\"): %s\n",
				remotepath, strerror(errno));
			if (errno == ECONNRESET)
				fget_disconnect_handle(ftp);
			return R_RETURN;
		}
	}

#ifdef DEBUG
	printf("    fget_download(): newpath = \"%s\"\n", newpath);
	printf("    fget_download(): localroot = \"%s\"\n", localroot);
#endif

	if (BIT_ISSET(options, OPT_CHASELINKS) && chase_l == NULL)
	{
		chase_l = fget_list_new(LIST_QUEUE, NULL);
		done_files_l = fget_list_new(LIST_QUEUE, NULL);
		if (chase_l == NULL || done_files_l == NULL)
		{
			fprintf(stderr, "fget: fget_list_new(): %s\n",
				strerror(errno));
			return R_RETURN;
		}
		freeflag = 1;
	}

	if (fget_stat(ftp, newpath, &fs) != R_FILEOK)
		return R_RETURN;

	if (S_ISDIR(fs.fs_mode))
	{
		if (!BIT_ISSET(options, OPT_HOSTDIRS))
		{
			if (strcmp(basename(remotepath), "/") != 0)
			{
				strlcat(localroot, "/", sizeof(localroot));
				strlcat(localroot, basename(remotepath),
					sizeof(localroot));
			}
			if (fget_mkdirhier(localroot, NULL, NULL) == -1)
			{
				fprintf(stderr,
					"fget: fget_mkdirhier(\"%s\"): %s\n",
					localroot, strerror(errno));
				return R_RETURN;
			}
		}
		retval = fget_recursion(ftp, newpath, fget_download_file,
					localroot);
		if (retval != R_RETURN
		    && BIT_ISSET(options, OPT_DELETELOCAL))
			fget_delete_local(ftp, newpath, localroot);
	}
	else
	{
		if (BIT_ISSET(options, OPT_HOSTDIRS)
		    && (cp = strrchr(localroot, '/')) != NULL)
			*cp = '\0';
		retval = fget_download_file(ftp, newpath, &fs, localroot);
	}

	if (freeflag && BIT_ISSET(options, OPT_CHASELINKS))
	{
		if (retval != R_RETURN)
		{
			fget_listptr_reset(&lp1);
			while (fget_list_next(chase_l, &lp1) != 0)
			{
				char *cp;

				cp = (char *)fget_listptr_data(&lp1);

				fget_listptr_reset(&lp2);
				if ((BIT_ISSET(options, OPT_HOSTDIRS) ||
				     strncmp(cp, newpath, strlen(newpath)) == 0)
				    && fget_list_search(done_files_l, &lp2, cp,
					(BIT_ISSET(options, OPT_RECURSIVE)
					 ? (fget_matchfunc_t)subdir_match
					 : NULL)) == 0
				    && fget_download(ftp, cp,
						     changeroot) == R_RETURN)
					break;
			}
		}
		fget_list_free(chase_l, free);
		fget_list_free(done_files_l, free);
		chase_l = NULL;
		done_files_l = NULL;
	}

	return retval;
}


