/*
**  Copyright 2000-2004 University of Illinois Board of Trustees
**  Copyright 2000-2004 Mark D. Roth
**  All rights reserved.
**
**  local.c - local file manipulation code for fget
**
**  Mark D. Roth <roth@feep.net>
*/

#include <fget.h>

#include <stdio.h>
#include <errno.h>
#include <utime.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <pwd.h>
#include <grp.h>

#ifdef STDC_HEADERS
# include <string.h>
# include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif


void
fget_set_perms(char *localpath, struct ftpstat *fsp)
{
	static fget_list_t *warned_users_l = NULL, *warned_groups_l = NULL;
	fget_listptr_t lp;
	struct passwd *pwd = NULL;
	struct group *grp = NULL;

	if (warned_users_l == NULL)
		warned_users_l = fget_list_new(LIST_QUEUE, NULL);
	if (warned_groups_l == NULL)
		warned_groups_l = fget_list_new(LIST_QUEUE, NULL);

	if (!S_ISLNK(fsp->fs_mode) && chmod(localpath, fsp->fs_mode) == -1)
		fprintf(stderr, "fget: %s: cannot change permissions: %s\n",
			localpath, strerror(errno));

	fget_listptr_reset(&lp);
	if (fget_list_search(warned_users_l, &lp, fsp->fs_username, NULL) == 0)
	{
		pwd = getpwnam(fsp->fs_username);
		if (pwd == NULL && verbose > 1)
		{
			fprintf(stderr, "fget: warning: no local account "
				"for user \"%s\"\n", fsp->fs_username);
			fget_list_add(warned_users_l,
				      strdup(fsp->fs_username));
		}
	}

	fget_listptr_reset(&lp);
	if (fget_list_search(warned_groups_l, &lp,
			     fsp->fs_groupname, NULL) == 0)
	{
		grp = getgrnam(fsp->fs_groupname);
		if (grp == NULL && verbose > 1)
		{
			fprintf(stderr, "fget: warning: no local account "
				"for group \"%s\"\n", fsp->fs_groupname);
			fget_list_add(warned_groups_l,
				      strdup(fsp->fs_groupname));
		}
	}

	if (pwd != NULL || grp != NULL)
	{
		if (
#ifdef HAVE_LCHOWN
			   lchown
#else
			   chown
#endif
			   (localpath, (pwd != NULL ? pwd->pw_uid : -1),
			    (grp != NULL ? grp->gr_gid : -1)) == -1)
			fprintf(stderr,
				"fget: %s: cannot change owner or group: %s\n",
				localpath, strerror(errno));
	}
}


static int
fget_extendfs(void)
{
	int status;

	if (extendfs_command != NULL)
	{
		printf("\nfget: executing command: \"%s\"\n",
		       extendfs_command);
		status = system(extendfs_command);
		if (WIFEXITED(status))
		{
			if (WEXITSTATUS(status) == 0)
				return 0;
			fprintf(stderr,
				"fget: system(): child exited with status %d\n",
				WEXITSTATUS(status));
		}
		else /* if (WIFSIGNALED(status)) */
			fprintf(stderr,
				"fget: system(): child died with signal %d\n",
				WTERMSIG(status));
		return 1;
	}

	return 0;
}


int
fget_download_dir(FTP* ftp, char *remotepath, struct ftpstat *fsp,
		  char *localpath, struct stat *sp)
{
	int ext_flag = 0;

#ifdef DEBUG
	printf("==> fget_download_dir(remotepath=\"%s\", localpath=\"%s\")\n",
	       remotepath, localpath);
#endif

	if (!S_ISDIR(sp->st_mode))
	{
		/* unlink the local file if it existed */
		if (sp->st_mode != 0 && fget_unlink(localpath, sp) == -1)
			return R_ERR;

		if (verbose)
			printf("fget: creating directory: %s\n", localpath);
		while (1)
		{
			if (mkdir(localpath, 0755) == -1)
			{
				if (!ext_flag && (errno == ENOSPC))
				{
					ext_flag = 1;
					if (fget_extendfs() == 0)
						continue;
				}
				fprintf(stderr, "fget: mkdir(\"%s\"): %s\n",
					localpath, strerror(errno));
				return (errno == ENOSPC ? R_RETURN : R_ERR);
			}
			ext_flag = 0;
			break;
		}

	}
	else if (verbose > 1)
	{
		printf("fget: directory already exists: %s\n", localpath);
		return R_SKIP;
	}

	return R_FILEOK;
}


int
fget_download_symlink(FTP* ftp, char *remotepath, struct ftpstat *fsp,
		      char *localpath, struct stat *sp,
		      char *linkto, size_t ltlen)
{
	int ext_flag = 0;
	char linkbuf[MAXPATHLEN];

	if (ftp_readlink(ftp, remotepath, linkbuf, sizeof(linkbuf)) == -1)
	{
		fprintf(stderr, "fget: ftp_readlink(\"%s\"): %s\n",
			remotepath, strerror(errno));
		return R_ERR;
	}

	if (BIT_ISSET(options, OPT_FIXLINKS) && linkbuf[0] == '/')
	{
		if (fget_relativepath(dirname(remotepath), linkbuf,
				 linkto, ltlen) == -1)
		{
			fprintf(stderr, "fget: fget_relativepath(): %s\n",
				strerror(errno));
			return R_ERR;
		}
	}
	else
		strcpy(linkto, linkbuf);

	/* check if the link already exists */
	if (S_ISLNK(sp->st_mode))
	{
		if (readlink(localpath, linkbuf, sizeof(linkbuf)) == -1)
		{
			fprintf(stderr, "fget: readlink(\"%s\"): %s\n",
				localpath, strerror(errno));
			return R_ERR;
		}
		linkbuf[sizeof(linkbuf) - 1] = '\0';
		if (strcmp(linkto, linkbuf) == 0)
		{
			if (verbose > 1)
				printf("fget: symlink already exists: %s\n",
				       localpath);
			return R_SKIP;
		}
	}

	/* unlink the local file if it exists */
	if (sp->st_mode != 0 && fget_unlink(localpath, sp) == -1)
		return R_ERR;

	if (verbose)
		printf("fget: creating symlink: %s -> %s\n", localpath,
		       linkto);
	while (1)
	{
		if (symlink(linkto, localpath) == -1)
		{
			if (!ext_flag && (errno == ENOSPC))
			{
				ext_flag = 1;
				if (fget_extendfs() == 0)
					continue;
			}
			fprintf(stderr, "fget: symlink(\"%s\", \"%s\"): %s\n",
				linkto, localpath, strerror(errno));
			return (errno == ENOSPC ? R_RETURN : R_ERR);
		}
		ext_flag = 0;
		break;
	}

	return R_FILEOK;
}


int
fget_download_regfile(FTP *ftp, char *remotepath, struct ftpstat *fsp,
		      char *localpath, struct stat *sp)
{
	char readbuf[FTPBUFSIZE];
	struct timeval before, after;
	float elapsed;
	int i, ofd = -1, ext_flag = 0;
	size_t sz;
	off_t bytecount = 0, nextreport = 0, interval = 10240;
	FTPFILE *ftpfile = NULL;

	/* check to see if the file already exists */
	if (sp->st_mode != 0
	    && (!BIT_ISSET(options, OPT_NEWER)
	        || sp->st_mtime >= fsp->fs_mtime)
	    && (!BIT_ISSET(options, OPT_DIFFSIZE)
		|| sp->st_size == fsp->fs_size))
	{
		if (verbose > 1)
			printf("fget: not updating file %s\n", localpath);
		return R_SKIP;
	}

	if (verbose)
		printf("fget: downloading: %s => %s...\n", remotepath,
		       localpath);

	if (ftp_open(&ftpfile, ftp, remotepath, O_RDONLY) == -1)
	{
		fprintf(stderr, "fget: ftp_open(\"%s\"): %s\n", remotepath,
			strerror(errno));
		return R_ERR;
	}

	while (1)
	{
		ofd = open(localpath, O_WRONLY | O_CREAT | O_TRUNC, 0644);
		if (ofd == -1)
		{
			if (!ext_flag && errno == ENOSPC)
			{
				ext_flag = 1;
				if (fget_extendfs() == 0)
					continue;
			}
			fprintf(stderr, "fget: open(\"%s\"): %s\n", localpath,
				strerror(errno));
			i = errno;
			if (ftp_close(ftpfile) == -1)
				fprintf(stderr, "fget: ftp_close(): %s\n",
					strerror(errno));
			errno = i;
			return (errno == ENOSPC ? R_RETURN : R_ERR);
		}
		ext_flag = 0;
		break;
	}

	elapsed = 0;
	gettimeofday(&before, NULL);
	while ((sz = ftp_read(ftpfile, readbuf, sizeof(readbuf))) > 0)
	{
		while (1)
		{
			if (write(ofd, readbuf, sz) == -1)
			{
				if (!ext_flag && errno == ENOSPC)
				{
					gettimeofday(&after, NULL);
					elapsed +=
					    (after.tv_sec - before.tv_sec) +
					    ((after.tv_usec -
					      before.tv_usec) / 1000000.0);
					ext_flag = 1;
					if (fget_extendfs() == 0)
					{
						gettimeofday(&before, NULL);
						continue;
					}
				}
				fprintf(stderr, "fget: write(): %s\n",
					strerror(errno));
				i = errno;
				if (ftp_close(ftpfile) == -1)
					fprintf(stderr,
						"fget: ftp_close(): %s\n",
						strerror(errno));
				errno = i;
				return (errno == ENOSPC ? R_RETURN : R_ERR);
			}
			ext_flag = 0;
			break;
		}
		bytecount += sz;
		if (verbose && isatty(fileno(stdout)) && bytecount > nextreport)
		{
			printf("    received %lu/%lu bytes\r",
			       (unsigned long)bytecount,
			       (unsigned long)fsp->fs_size);
			fflush(stdout);
			nextreport += interval;
		}
	}
	gettimeofday(&after, NULL);
	elapsed += (after.tv_sec - before.tv_sec) +
	    ((after.tv_usec - before.tv_usec) / 1000000.0);
	if (verbose)
		printf("    received %lu bytes in %.2f seconds (%.2f K/s)\n",
		       (unsigned long)bytecount, elapsed,
		       (((float)bytecount / 1024.0) / elapsed));

	if (ftp_close(ftpfile) == -1)
	{
		fprintf(stderr, "fget: ftp_close(): %s\n", strerror(errno));
		return R_ERR;
	}

	if (close(ofd) == -1)
	{
		fprintf(stderr, "fget: close(): %s\n", strerror(errno));
		return R_ERR;
	}

	return R_FILEOK;
}


